#!/usr/bin/env python

"""Zamienia treebank na JSON do wrzucenia do datasetu

"""

import argparse
import gzip
import json
from lxml import etree
from pathlib import Path
from skladnica_trees import ConstituencyTree
import sys


def mark_dependencies(node, hd_id=None, deprel='root'):
    if node.is_terminal():
        node.features['hd_id'] = hd_id
        node.features['deprel'] = deprel
        return node.tid
    else:
        head = [n for n in node.children if n.is_head][0]
        my_center = mark_dependencies(head, hd_id, deprel)
        for c in node.children:
            if c.is_head:
                continue
            mark_dependencies(c, my_center, c.deprel)
        return my_center


def generate_nonterminal(node):
    if node.is_terminal():
        return {
            'cat': None,
            'children': [node.tid],
        }
    else:
        return {
            'cat': node.cat,
            'children': [n.nid for n in node.children],
        }


all_tags = set()
all_deprels = set()
parser = etree.XMLParser(remove_blank_text=True)


def process_treebank_file(tree_file):
    tree_xml = json.load(tree_file.open())
    tree = ConstituencyTree.from_JSON_dict(tree_xml)
    # tree_xml = etree.parse(tree_file.open(), parser)
    # tree = ConstituencyTree.from_XML(tree_xml)
    md = tree.metadata
    # split_idx = md['split']
    if 'corp_id' not in md:
        md['corp_id'] = 'Tiger'  # !!!
    split_idx = md['split']
    # Tokens:
    for i, t in enumerate(tree.tokens):
        t.tid = i
    mark_dependencies(tree.root)
    tokens = [(t.orth, t.lemma, t.pos, t.tag,
               t.features['hd_id'], t.features['deprel'],)
              for t in tree.tokens]
    token_dict = dict(zip(['tokens', 'lemmas',
                           'poses', 'tags',
                           'heads', 'deprels'],
                          zip(*tokens)))
    all_tags.update(token_dict['tags'])
    all_deprels.update(token_dict['deprels'])
    # Nonterminals:
    for i, n in enumerate(tree.root.traverse_preorder()):
        n.nid = i
    nodes = [generate_nonterminal(n) for n in tree.root.traverse_preorder()]
    # Assemble output:
    print('🔶', md['sent_id'], file=sys.stderr)
    json_dict = {}
    json_dict["corp_id"] = md['corp_id']
    json_dict["sent_id"] = f"Tiger/{md['text_id']}/{md['sent_id']}"
    json_dict.update(token_dict)
    json_dict["nonterminals"] = nodes
    return json_dict, split_idx, len(token_dict["tokens"])


split_len = {'dev': 0, 'test': 0, 'train': 0, }


def process_treebank_files(treebank, dataset_dir):
    Path(dataset_dir).mkdir(parents=True, exist_ok=True)
    with (gzip.open(dataset_dir+'/treebank_dev.jsonl.gz', 'wt') as devf,
          gzip.open(dataset_dir+'/treebank_test.jsonl.gz', 'wt') as testf,
          gzip.open(dataset_dir+'/treebank_train.jsonl.gz', 'wt') as trainf):
        split_fs = {'dev': devf, 'test': testf, 'train': trainf, }
        corp_path = Path(treebank)
        for tree_file in sorted(corp_path.glob('**/*.json')):
            # print(tree_file, file=sys.stderr)
            json_dict, split_idx, sent_len = process_treebank_file(tree_file)
            split_len[split_idx] += sent_len
            print(json.dumps(json_dict, ensure_ascii=False),
                  file=split_fs[split_idx])


argparser = argparse.ArgumentParser(description=__doc__)
argparser.add_argument('treedir', help="directory containing the treebank")
argparser.add_argument('dsdir', help="output directory")
args = argparser.parse_args()

process_treebank_files(args.treedir, args.dsdir)
print(split_len, file=sys.stderr)
for t in sorted(all_tags):
    print(f"                                '{t}',")
