from lxml import etree


class ConstituencyTree:

    def __init__(self, metadata, tree, tokens):
        self.metadata = metadata
        self.root = tree
        self.tokens = tokens

    @staticmethod
    def _node_from_XML(node):
        span_s = int(node.get('from'))
        span_e = int(node.get('to'))
        assert span_s is not None
        nonterminal = node.find('nonterminal')
        if nonterminal is None:
            terminal = node.find('terminal')
            assert terminal is not None
            orth = terminal.find('orth').text
            lemma = terminal.find('base').text
            tag = terminal.find('f[@type="tag"]').text
            attributes = dict(terminal.attrib)
            return {'orth': orth,
                    'lemma': lemma,
                    'tag': tag,
                    'span': (span_s, span_e),
                    'features': attributes,
                    }
        else:
            cat = nonterminal.find('category').text
            features = {f.get('type'): f.text
                        for f in nonterminal.findall('f')}
            children = node.findall('children/child')
            child_ids = [(c.get('nid'), c.get('deprel'),
                          c.get('head') == "true")
                         for c in children]
            return {'cat': cat,
                    'features': features,
                    'span': (span_s, span_e),
                    'children': child_ids,
                    }

    @classmethod
    def _make_tree(cls, nid, deprel, is_head, parent, nodes, tokens):
        node = nodes[nid]
        node['is_head'] = is_head
        node['deprel'] = deprel
        if 'cat' in node:
            tree_node = NonterminalNode(
                parent, node['span'], node['is_head'], node['deprel'],
                node['cat'], node['features'])
            children = [cls._make_tree(c, d, h, tree_node, nodes, tokens)
                        for c, d, h in node['children']]
            tree_node.children = children
        else:
            tree_node = TerminalNode(
                parent, node['span'], node['is_head'], node['deprel'],
                node['orth'], node['lemma'], node['tag'],
                node['features'])
            tokens.append(tree_node)
        return tree_node

    @classmethod
    def from_XML(cls, tree_xml):
        metadata = dict(tree_xml.getroot().attrib)
        text = tree_xml.find('text').text
        if text:
            metadata['text'] = text
        nodes = {}
        for node in tree_xml.findall('node'):
            nid = node.get('nid')
            nodes[nid] = cls._node_from_XML(node)
        assert all(i == int(n) for i, n in enumerate(nodes))
        tokens = []
        root = cls._make_tree(nid='0', deprel='hd',
                              is_head=True, parent=None,
                              nodes=nodes, tokens=tokens)
        tokens.sort(key=lambda t: (int(t.span[0]), int(t.span[1])))
        return cls(metadata, root, tokens)

    def to_XML(self):
        """Returns an lxml.etree object representing the tree
        """
        xml_tree = etree.Element('forest')
        text = self.metadata.pop('text')
        for key in self.metadata:
            xml_tree.set(key, self.metadata[key])
        etree.SubElement(xml_tree, 'text').text = text
        startnode = etree.SubElement(xml_tree, 'startnode')
        startnode.text = self.root.cat
        startnode.set('from', str(self.root.span[0]))
        startnode.set('to', str(self.root.span[1]))
        answer = etree.SubElement(xml_tree, 'answer-data')
        etree.SubElement(answer, 'base-answer', type="FULL",
                         username="forgotten")
        # number the nodes:
        nodes = []
        for id, node in enumerate(self.root.traverse_preorder()):
            node.nid = id
            nodes.append(node)
        for node in nodes:
            xml_tree.append(node.to_XML())
        return xml_tree

    @classmethod
    def from_JSON_dict(cls, json_dict):
        metadata = json_dict['metadata'].copy()
        tokens = []
        root = TreeNode.from_JSON_dict(json_dict['tree'], None, tokens)
        tokens.sort(key=lambda t: (int(t.span[0]), int(t.span[1])))
        return cls(metadata, root, tokens)

    def to_JSON_dict(self):
        """Returns a JSON-serialisable dict representing the tree
        """
        json_dict = {
            'metadata': self.metadata,
            'tree': self.root.to_JSON_dict(),
        }
        return json_dict

    def mark_dependencies_on_tokens(self):
        """Add dependency heads and relations as direct atttributes of
        tokens making them easier to access e.g. for conversion to
        CONLL-like formats.

        """
        for i, t in enumerate(self.tokens):
            t.tid = i
        self.root._mark_dependencies(None, 'root')


class TreeNode:

    @classmethod
    def from_JSON_dict(cls, json_dict, parent, tokens):
        span = json_dict['span']
        if 'category' in json_dict:
            nt = NonterminalNode(parent,
                                 (span['from'], span['to']),
                                 json_dict['is_head'],
                                 json_dict.get('deprel'),
                                 json_dict['category'],
                                 json_dict.get('attributes', {}))
            nt.children = [cls.from_JSON_dict(c, nt, tokens)
                           for c in json_dict['children']]
            return nt
        else:
            tn = TerminalNode(parent,
                              (span['from'], span['to']),
                              json_dict['is_head'],
                              json_dict.get('deprel'),
                              json_dict['orth'],
                              json_dict['base'],
                              json_dict['tag'],
                              json_dict.get('features', {}))
            tokens.append(tn)
            return tn

    def __init__(self, parent, span, is_head, deprel):
        self.parent = parent
        self.span = span
        self.is_head = is_head
        self.deprel = deprel
        self.children = []

    def set_as_head(self):
        '''Mark self as the head child of its parent'''
        for c in self.parent.children:
            c.is_head = False
        self.is_head = True

    def get_head_child(self):
        hc = [c for c in self.children if c.is_head]
        assert len(hc) == 1
        return hc[0]

    def add_parent(self, cat, features):
        '''Introduce a new nonterminal with cat and features
        as the new parent of self'''
        new_node = NonterminalNode(self.parent, self.span,
                                   self.is_head, self.deprel, cat, features)
        idx = self.parent.children.index(self)
        self.parent.children[idx] = new_node
        self.parent = new_node
        self.is_head = True
        new_node.children = [self]

    def traverse_postorder(self):
        for c in self.children:
            yield from c.traverse_postorder()
        yield self

    def traverse_preorder(self):
        yield self
        for c in self.children:
            yield from c.traverse_preorder()

    def _update_span(self):
        if self.is_terminal():
            return
        new_span = (min(c.span[0] for c in self.children),
                    max(c.span[1] for c in self.children))
        if new_span != self.span:
            self.span = new_span
            if self.parent is not None:
                self.parent._update_span()

    def reparent(self, new_parent, order_by_span=False):
        '''Make self a child of new_parent

        The children are ordered by their central tokens. This
        produces a nice image of the tree in the case of discontiguous
        structures. Alternatively the children may be ordered by their
        spans (span starts in fact), which allows to work with trees
        missing some heads, but may result in sub-optimal child order.

        '''
        if len(self.parent.children) < 2:
            raise RuntimeError('Cannot reparent the only child')
        # remove from current parent:
        idx = self.parent.children.index(self)
        self.parent.children.pop(idx)
        # insert at new parent:
        if order_by_span:
            our_idx = self.span[0]
        else:
            our_idx = self.find_central_terminal().span[0]
        for i, c in enumerate(new_parent.children):
            if order_by_span:
                c_idx = c.span[0]
            else:
                c_idx = c.find_central_terminal().span[0]
            if c_idx > our_idx:
                break
        else:
            i = len(new_parent.children) + 1
        new_parent.children.insert(i, self)
        # The order is important here:
        new_parent._update_span()
        self.parent._update_span()
        self.parent = new_parent


class NonterminalNode(TreeNode):
    def __init__(self, parent, span, is_head, deprel, cat, features):
        super().__init__(parent, span, is_head, deprel)
        self.cat = cat
        self.features = features
        self.children = []

    def is_terminal(self):
        return False

    def __repr__(self):
        dep = f'({self.deprel})' if self.deprel is not None else ''
        repr_children = ','.join(repr(c) for c in self.children)
        return (f"{'*' if self.is_head else dep}"
                f"{self.cat}<{repr_children}>")

    def find_central_terminal(self):
        for c in self.children:
            if c.is_head:
                return c.find_central_terminal()
        raise RuntimeError('Search for head in a headless nonterminal')

    def get_text(self):
        return ''.join(c.get_text() for c in self.children)

    def get_tokens(self):
        toks = []
        for c in self.children:
            toks.extend(c.get_tokens())
        return toks

    def remove_from_tree(self):
        if self.parent is None:
            raise RuntimeError(f"Cannot remove the root node: {self}")
        idx = self.parent.children.index(self)
        self.parent.children.pop(idx)
        for c in reversed(self.children):
            c.parent = self.parent
            if c.is_head:
                c.is_head = self.is_head
                c.deprel = self.deprel
            self.parent.children.insert(idx, c)

    def to_XML(self):
        node = etree.Element('node', nid=str(self.nid))
        node.set('from', str(self.span[0]))  # from is reserved keyword
        node.set('to', str(self.span[1]))
        node.set('chosen', "true")
        nt = etree.SubElement(node, 'nonterminal')
        etree.SubElement(nt, 'category').text = self.cat
        for feat in self.features:
            etree.SubElement(nt, 'f', type=feat).text = str(self.features[feat])
        children = etree.SubElement(node, 'children', chosen="true")
        for c in self.children:
            child = etree.SubElement(children, 'child', nid=str(c.nid))
            child.set('from', str(c.span[0]))
            child.set('to', str(c.span[1]))
            child.set('head', "true" if c.is_head else "false")
            if c.deprel is not None:
                child.set('deprel', c.deprel)
        return node

    def to_JSON_dict(self):
        children = [c.to_JSON_dict() for c in self.children]
        nonterminal = {
            "category": self.cat,
            "is_head": self.is_head,
            "span": {
                "from": int(self.span[0]),
                "to": int(self.span[1]),
            },
            "attributes": self.features.copy(),
            "children": children,
        }
        if self.deprel is not None:
            nonterminal['deprel'] = self.deprel
        return nonterminal

    def _mark_dependencies(self, hd_id, deprel):
        if deprel is None:
            deprel = 'dep'
        head = [n for n in self.children if n.is_head][0]
        my_center = head._mark_dependencies(hd_id, deprel)
        for c in self.children:
            if c is head:
                continue
            c._mark_dependencies(my_center, c.deprel)
        return my_center


class TerminalNode(TreeNode):
    def __init__(self, parent, span, is_head, deprel, orth, base, tag,
                 features):
        super().__init__(parent, span, is_head, deprel)
        self.orth = orth
        self.lemma = base
        self.tag = tag
        self.features = features

    def is_terminal(self):
        return True

    @property
    def pos(self):
        return self.tag.split(':')[0]

    @property
    def cat(self):
        return None

    def __repr__(self):
        dep = f'({self.deprel})' if self.deprel is not None else ''
        return ('*' if self.is_head else dep) + \
            f"[{self.lemma}:{self.pos}]"

    def find_central_terminal(self):
        return self

    def get_text(self):
        return ('' if self.features.get('nps', None) == 'true'
                else ' ') + self.orth

    def get_tokens(self):
        return [self]

    def remove_from_tree(self):
        raise RuntimeError(f"Cannot remove a terminal: {self}")

    @staticmethod
    def _serialize_feature(f):
        if isinstance(f, bool):
            return "true" if f else "false"
        elif isinstance(f, str):
            return f
        else:
            return str(f)

    def to_XML(self):
        node = etree.Element('node', nid=str(self.nid))
        node.set('from', str(self.span[0]))  # from is a keyword
        node.set('to', str(self.span[1]))
        node.set('chosen', "true")
        terminal = etree.SubElement(node, 'terminal')
        etree.SubElement(terminal, 'orth').text = self.orth
        etree.SubElement(terminal, 'base').text = self.lemma
        etree.SubElement(terminal, 'f', type='tag').text = self.tag
        for key in self.features:
            terminal.set(key, self._serialize_feature(self.features[key]))
        terminal.set("disamb", "true")
        return node

    def to_JSON_dict(self):
        terminal = {
            "is_head": self.is_head,
            "span": {
                "from": int(self.span[0]),
                "to": int(self.span[1]),
            },
            "orth": self.orth,
            "base": self.lemma,
            "tag": self.tag,
            'features': self.features.copy(),
        }
        if self.deprel is not None:
            terminal['deprel'] = self.deprel
        return terminal

    def _mark_dependencies(self, hd_id, deprel):
        if deprel is None:
            deprel = 'dep'
        self.features['hd_id'] = hd_id
        self.features['deprel'] = deprel
        return self.tid
